import collections
import random
from enum import *



class TreeNode:
    def __init__(self, language_element, subnodes):
        self.language_element = language_element
        self.subnodes = subnodes

    def __str__(self):
        return str(self.language_element) + " " + str(self.subnodes)


def walk_the_tree(tree, level = 1):
    if tree == None:
        return

    print("    " * (level -1) + "|---" + str(tree.language_element) + " with " + str(len(tree.subnodes)) + " subnodes")
    for subnode in tree.subnodes:
        walk_the_tree(subnode, level + 1)




CFG = collections.namedtuple("CFG", "nonterminals terminals rules start")
# We define a context-free grammar as:
# 1. A list of nonterminal symbols
# 2. A list of terminal symbols
# 3. A list of production rules
# 4. A start symbol.

# Each symbol is defined by a positive integer.


# For the ETF expression grammar:

# Terminal symbols are:
# INTEGER, ADDOP, MULTOP, OPENPAREN, CLOSEPAREN

# Nonterminal symbols are:
# EXPRESSION, TERM, FACTOR

# The start symbol is EXPRESSION

# The rules are:

# EXPRESSION -> EXPRESSION ADDOP TERM
# EXPRESSION -> TERM
# TERM -> TERM MULTOP FACTOR
# TERM -> FACTOR
# FACTOR -> OPENPAREN EXPRESSION CLOSEPAREN
# FACTOR -> INTEGER

symbols = Enum("Symbols", "EXPRESSION TERM FACTOR INTEGER ADDOP MULTOP OPENPAREN CLOSEPAREN")
start = symbols.EXPRESSION



# The rules are a list of tuples, each of which represents a rule, as follows:

# (Nonterminal symbol, [ordered list of symbols (terminal or nonterminal) that are reduced to aforementioned nonterminal])

rules = [
         (symbols.EXPRESSION, [symbols.EXPRESSION, symbols.ADDOP,      symbols.TERM      ]),
         (symbols.EXPRESSION, [symbols.TERM                                              ]),
         (symbols.TERM,       [symbols.TERM,       symbols.MULTOP,     symbols.FACTOR    ]),
         (symbols.TERM,       [symbols.FACTOR                                            ]),
         (symbols.FACTOR,     [symbols.OPENPAREN,  symbols.EXPRESSION, symbols.CLOSEPAREN]),
         (symbols.FACTOR,     [symbols.INTEGER                                           ])
         ]

#print(rules)



# Once we have a CFG's definition, we can use Boltzmann sampling in order to generate random strings
# from that CFG.


# This, however, is not Boltzmann sampling.

# We examine all the leaves on the current derivation tree. Those mark where we possibly could apply a rule.
# We then determine which leaves are amenable to rule application, and then we randomly select a leaf and for
# that leaf, we randomly select a rule, and apply it.

# To avoid being accidentally quadratic, we cache a list of the leaves, and also cache the current length of
# the generated string (in order to do pseudo-Boltzmann sampling).


def indexes_to_item(tree, location):
    opsite = tree
    if (location != []):
        for depth_index in location:
            opsite = opsite.subnodes[depth_index]
            #print("TMP index ", depth_index, " gives ",opsite)
    else:
        opsite = tree

    return opsite

def random_leaf_random_rule(tree, leafcache, ruleset):

    for x in range(16):
        proband_cache_index = random.randrange(len(leafcache))
        location = leafcache[proband_cache_index]

        opsite = indexes_to_item(tree, location)

        proband_rule_index = random.randrange(len(ruleset))
        rule = ruleset[proband_rule_index]

        if (opsite.language_element == rule[0]):
            return (proband_cache_index, proband_rule_index)


    # we failed somehow, just return None
    return None






def apply_rule(tree, leafcache, cache_index, ruleset, rule_number):
    # first, we isolate the place on the tree that we are going to do the graft.
    opsite = tree

    location = leafcache[cache_index]

    opsite = indexes_to_item(tree, location)

    #print(opsite)
    #walk_the_tree(opsite)

    # now we verify that we can legitimately apply the given rule here.

    the_rule = ruleset[rule_number]

#    print("We will try applying the following rule:", the_rule)

    # We verify that the item at the operation site is:
    # 1) a leaf (it has no subnodes)
    # and
    # 2) has the correct value

#    print("Opsite language value", opsite.language_element)
#    print("Rule expects", the_rule[0])

    assert(opsite.language_element == the_rule[0])

    assert(opsite.subnodes == [])

#    print("opsite has no subnodes, all good")

    new_leaves = []

    for idx, new_subnode in enumerate(the_rule[1]):
        #print(new_subnode)
        x = TreeNode(new_subnode, [])
        opsite.subnodes.append(x)

        leafcache.append(location + [idx])

#    print("Now the opsite looks like")
#    walk_the_tree(opsite)

    # update the leaf cache

    del leafcache[cache_index]



#def next_stage(derivation_tree_with_cached, language_definition):




def the_tree_farm():
    for x in range(16):
        selected = random_leaf_random_rule(test_tree, leafcache, rules)
        if (selected == None):
            # we have achieved what we need! all leaves are terminal items
            return test_tree
        else:
            (next_index_cache, next_rule) = selected

#        print ("next cache index is", next_index_cache, "next rule is", next_rule)
        apply_rule(test_tree, leafcache, next_index_cache, rules, next_rule)

#        print("and the tree looks like")
        #walk_the_tree(test_tree)

#        print("updated leafcache")
#        print(leafcache)

#


test_tree = TreeNode(symbols.EXPRESSION, [])

#walk_the_tree(test_tree)


def tree_to_sentence(tree_root, leaves):
    for node in tree_root.subnodes:
        if node.subnodes == []:
            leaves.append(node.language_element)
        else:
            tree_to_sentence(node, leaves)




if __name__ == '__main__':

    leafcache = [[]]

    while True:
        produced_tree = the_tree_farm()
        if(produced_tree != None):
            print("SUCCESS!!!!")
            walk_the_tree(produced_tree)
            print("now to try to extract the sentence,,")
            leaves = []
            tree_to_sentence(produced_tree, leaves)
            print(leaves)
            break








# Deterministic tree creation. W





# Actual Boltzmann sampling goes here.

# We use the "sandwich" technique in https://notebook.drmaciver.com/posts/2020-07-11-10:49.html
# to estimate the Boltzmann distribution generating function. However, to enumerate the number of
# words in the grammar with length n, we use the technique described in 
#
# "Generating Strings at Random from a Context Free Grammar" by Bruce McKenzie, 1997

# The notation is therefore as follows from that paper:

# The paper uses the notation ||x||_n to mean "the number of strings of length n generated by
# the symbol x", but since this is text and not LaTeX, we use the notation "||X||==n" to mean
# the same.

# Nonterminals are indexed by i, i = [1, number of nonterminals]
# Each nonterminal may have multiple ways of generating it. For each nonterminal i,
# there may be multiple rules that create it. For each i, there is an index j of these
# multiple ways.

# Production rules are thus indexed by:
# (i = [1, number of nonterminals], j = [1, number of ways to generate nonterminal i]).

# Each production rule is represented by α_ij = x_ij1, x_ij2, ..., x_ijT, with T=T_ij
# representing the number of symbols in the reduce rule (i, j)

# We have two functions, and we use memoization in order to evaluate them without
# pathological cost.


# The first function, f(i, N), which we will call Fzero, represents the number
# of strings with length exactly N that can be generated by the expansion of
# nonterminal number i. It is a list of integers, with one element for each 
# production rule that generates nonterminal i

# f(i, N) = [||α_ij||==N for j = [1, number of production rules for nonterminal i]]


# The second function, f'(i, j, k, N), which we will call Fprim, represents the
# number of strings with length exactly N that can be generated by the
# last k symbols of the specific production rule (i, j):

# f'(i, j, k, N) 













# Furthermore, we also note that the description of a context-free grammar is *itself* context-free
# so if we take the CFG-description grammar (BNF or something isomorphic to it), and use Boltzmann
# sampling on *it*, we will generate candidate grammars; which will be used to test the FPGA-based
# parser generator against a reference LR parser generator implementation. Naturally, this procedure 
# may not produce LR-parseable grammars or even deterministic context-free grammars; but if the grammar
# is rejected by both the FPGA parser generator and reference parser generator; this is fine, we just
# move on to the next one.


# Once we randomly generate a CFG, we randomly generate strings that are valid strings, but also
# generate exemplars that are potentially *invalid* strings (both by mutating valid strings but
# also by generating random sequences ab initio from the terminal symbols). The potentially invalid
# strings are tested against the reference parser, those that are accidentally valid are discarded,
# and those that are invalid are used in the automatic equivalence testing.


# We fail our candidate parser generator if:
#
# 1. During the generation phase, it rejects a grammar the reference generator doesn't reject, or
#    accepts a grammar the reference generator rejects.

# 2. During the execution phase, it rejects a string the reference generated parser doesn't reject,
#    or accepts a string the reference generated parser rejects.

# 3. During the execution phase, it accepts a string that the reference generated parser accepts; but
#    the parse tree it generates is not identical to the parse tree that the reference generated parser
#    produced.



